#lang sicp

(#%require racket/trace)
(#%require (only racket/base when))

(#%require "table.rkt")

(define (make-dispatch-table types-procs-alist)
  (let ((eval-dispatch-table (make-1d-table)))
    (define (eval-dispatch-put! type proc)
      ((eval-dispatch-table 'insert-proc!) type proc))
    (define (eval-dispatch-lookup type)
      ((eval-dispatch-table 'lookup-proc) type))
    (define (build-table types-procs)
      (if (pair? types-procs)
          (let ((type (caar types-procs))
                (proc (cadar types-procs)))
            (begin (eval-dispatch-put! type proc)
                   (build-table (cdr types-procs))))))
    (define (dispatch m)
      (cond ((eq? m 'lookup) eval-dispatch-lookup)
            (else (error "Unknown input - DISPATCH-TABLE" m))))
    (build-table types-procs-alist)
    dispatch))


;; This is a lightly modified version of ch4-mceval.scm to work in racket.

;;;;METACIRCULAR EVALUATOR FROM CHAPTER 4 (SECTIONS 4.1.1-4.1.4) of
;;;; STRUCTURE AND INTERPRETATION OF COMPUTER PROGRAMS

;;;;Matches code in ch4.scm

;;;;This file can be loaded into Scheme as a whole.
;;;;Then you can initialize and start the evaluator by evaluating
;;;; the two commented-out lines at the end of the file (setting up the
;;;; global environment and starting the driver loop).

;;;;**WARNING: Don't load this file twice (or you'll lose the primitives
;;;;  interface, due to renamings of apply).

;;;from section 4.1.4 -- must precede def of metacircular apply
;; Racket doesn't seem to like this.  Instead rename the apply here to
;; mce-apply.
;;(define apply-in-underlying-scheme apply)

;;;SECTION 4.1.1

(#%provide (all-defined))

(define type-tag car)

;; Data directed eval. Consult the eval dispatch table that uses the
;; car of the expression to get the type-tag.  Only if nothing matches from
;; the dispatch table, attempt to evaluate as an application.  This is very similar
;; to the data-directed dispatch for symbolic differentiation in exercise
;; 3.73.  The main difference is that we have another case to consider
;; if the lookup fails (ie application).
(define (eval exp env)
  (cond ((self-evaluating? exp) exp)
        ((variable? exp) (lookup-variable-value exp env))
        ((eval-dispatch-lookup (type-tag exp))
         => (lambda (proc) (proc exp env)))
        ((application? exp)
         (mce-apply (eval (operator exp) env)
                (list-of-values (operands exp) env)))
        (else
         (error "Unknown expression type -- EVAL" exp))))

(define (mce-apply procedure arguments)
  (cond ((primitive-procedure? procedure)
         (apply-primitive-procedure procedure arguments))
        ((compound-procedure? procedure)
         (eval-sequence
           (procedure-body procedure)
           (extend-environment
             (procedure-parameters procedure)
             arguments
             (procedure-environment procedure))))
        (else
         (error
          "Unknown procedure type -- MCE-APPLY" procedure))))


(define (list-of-values exps env)
  (if (no-operands? exps)
      '()
      (cons (eval (first-operand exps) env)
            (list-of-values (rest-operands exps) env))))

(define (eval-if exp env)
  (if (true? (eval (if-predicate exp) env))
      (eval (if-consequent exp) env)
      (eval (if-alternative exp) env)))

(define (eval-sequence exps env)
  (cond ((last-exp? exps) (eval (first-exp exps) env))
        (else (eval (first-exp exps) env)
              (eval-sequence (rest-exps exps) env))))

(define (eval-assignment exp env)
  (set-variable-value! (assignment-variable exp)
                       (eval (assignment-value exp) env)
                       env)
  'ok)

(define (eval-definition exp env)
  (define-variable! (definition-variable exp)
                    (eval (definition-value exp) env)
                    env)
  'ok)

(define (eval-unbind exp env)
  (make-unbound-variable! (unbound-variable exp) env))

;;;SECTION 4.1.2

(define (self-evaluating? exp)
  (cond ((number? exp) true)
        ((string? exp) true)
        (else false)))

(define (text-of-quotation exp) (cadr exp))

(define (tagged-list? exp tag)
  (if (pair? exp)
      (eq? (car exp) tag)
      false))

(define (variable? exp) (symbol? exp))

;; Assignment
(define (assignment-variable exp) (cadr exp))
(define (assignment-value exp) (caddr exp))
(define (make-assignment variable value)
  (list 'set! variable value))

;; Definition
(define (definition-variable exp)
  (if (symbol? (cadr exp))
      (cadr exp)
      (caadr exp)))
(define (definition-value exp)
  (if (symbol? (cadr exp))
      (caddr exp)
      (make-lambda (cdadr exp)
                   (cddr exp))))

;; Unbinding
(define (make-unbound? exp) (tagged-list? exp 'make-unbound!))
(define (unbound-variable exp) (cadr exp))

;; Lambda
(define (lambda-parameters exp) (cadr exp))
(define (lambda-body exp) (cddr exp))
(define (make-lambda parameters body)
  (cons 'lambda (cons parameters body)))

;; if
(define (if-predicate exp) (cadr exp))
(define (if-consequent exp) (caddr exp))
(define (if-alternative exp)
  (if (not (null? (cdddr exp)))
      (cadddr exp)
      'false))
(define (make-if predicate consequent alternative)
  (list 'if predicate consequent alternative))

;; Sequence
(define (begin-actions exp) (cdr exp))

(define (last-exp? seq) (null? (cdr seq)))
(define (first-exp seq) (car seq))
(define (rest-exps seq) (cdr seq))

(define (sequence->exp seq)
  (cond ((null? seq) seq)
        ((last-exp? seq) (first-exp seq))
        (else (make-begin seq))))

(define (make-begin seq) (cons 'begin seq))

;; Application
(define (application? exp) (pair? exp))
(define (operator exp) (car exp))
(define (operands exp) (cdr exp))
(define (make-application operator operands)
  (cons operator operands))

(define (no-operands? ops) (null? ops))
(define (first-operand ops) (car ops))
(define (rest-operands ops) (cdr ops))
(define (last-operand? ops) (null? (rest-operands ops)))

;; cond (including recipient clause for exercise 4.5)
(define (cond-clauses exp) (cdr exp))
(define (cond-else-clause? clause)
  (eq? (cond-predicate clause) 'else))
(define (cond-recipient-clause? clause)
  (eq? (cond-infix-op clause) '=>))
(define (cond-predicate clause) (car clause))
(define (cond-infix-op clause) (cadr clause))
(define (cond-recipient clause) (caddr clause))
(define (cond-actions clause) (cdr clause))

(define (cond->if exp)
  (expand-clauses (cond-clauses exp)))

(define (expand-clauses clauses)
  (if (null? clauses)
      'false                          ; no else clause
      (let ((first (car clauses))
            (rest (cdr clauses)))
        (cond ((cond-else-clause? first)
               (if (null? rest)
                   (sequence->exp (cond-actions first))
                   (error "ELSE clause isn't last -- COND->IF"
                          clauses)))
              ((cond-recipient-clause? first)
               (let ((test (cond-predicate first))
                     (proc (cond-recipient first)))
                 (if proc
                     (make-if test
                              (make-application proc test)
                              (expand-clauses rest))
                     (error "RECIPIENT clause missing -- COND->IF"))))
              (else (make-if (cond-predicate first)
                             (sequence->exp (cond-actions first))
                             (expand-clauses rest)))))))


;; Exercise 4.4
;; and / or
;; Truth expressions are a list whose car is 'and or 'or and where
;; the rest of the list is the expressions to evaluate.
(define (first-conjunct conjuncts) (cadr conjuncts))
(define (rest-conjuncts conjuncts) (cddr conjuncts))
(define (last-conjunct? conjuncts) (null? (rest-conjuncts conjuncts)))

(define (make-and conjuncts) (cons 'and conjuncts))

;; As special forms
(define (eval-and conjuncts env)
  (cond ((null? conjuncts)
         'true)
        ((last-conjunct? conjuncts)
         (eval (first-conjunct conjuncts) env))
        ((false? (eval (first-conjunct conjuncts) env))
         'false)
        (else
         (eval-and (make-and (rest-conjuncts conjuncts)) env))))

(define (first-disjunct disjuncts) (cadr disjuncts))
(define (rest-disjuncts disjuncts) (cddr disjuncts))
(define (last-disjunct? disjuncts) (null? (rest-disjuncts disjuncts)))

(define (make-or disjuncts) (cons 'or disjuncts))

(define (eval-or disjuncts env)
  (if (null? disjuncts)
      'false
      (let ((val (eval (first-disjunct disjuncts) env)))
        (if val
            val
            (eval-or (make-or (rest-disjuncts disjuncts)) env)))))

;; As derived forms.  Not complete; would need negation, which we don't
;; seem to have yet.
;;
;; (and a b) -> (if (eq? a #f) #f (if (eq? b #f) #f b))
;; (or a b)  -> (if a a (if b b #f))

;; (define (and->if predicates)
;;   (expand-and-to-if predicates))

;; (define (expand-and-to-if predicates)
;;   (if (null? predicates)
;;       'true
;;       (let ((first (first-predicate predicates))
;;             (rest (rest-predicates predicates)))
;;         (if (last-predicate? predicates)
;;             first
;;             (make-if (false? first)     ; Need an expression not a value
;;                      'false
;;                      (expand-and-to-if rest))))))

;; Exercise 4.6

;; Reduce (let ((var1 exp1) ... (varn expn)) body) to
;; ((lambda (var1 ... varn) (body)) (exp1 ... expn))

(define (let? exp) (tagged-list? exp 'let))
(define (named-let? exp) (and (let? exp) (not (pair? (cadr exp)))))

(define (named-let-var exp) (cadr exp))
(define (let-clauses exp) (if (named-let? exp) (caddr exp) (cadr exp)))
(define (let-body exp) (if (named-let? exp) (cdddr exp) (cddr exp)))
(define (let-clauses-first clauses) (car clauses))
(define (let-clauses-rest clauses) (cdr clauses))
(define (let-clause-last? clauses) (null? (let-clauses-rest clauses)))
(define (let-clause-var clause) (car clause))
(define (let-clause-exp clause) (cadr clause))
(define make-body list)
(define (make-let clauses body)
  (cons 'let (cons clauses body)))
(define (make-named-let var clauses body)
  (cons 'let (cons var (cons clauses body))))

(define (let->combination exp)
  (let ((clauses (let-clauses exp))
        (body (let-body exp)))
    (let ((clause-vars (map let-clause-var clauses))
          (clause-exps (map let-clause-exp clauses)))
      (if (named-let? exp)
          (let ((named-proc (let-clause-exp exp)))
            (make-application
             (make-lambda
              `(,named-proc)
              (make-body
               (make-application
                (make-lambda
                 '(proc)
                 (make-body
                  (make-assignment named-proc 'proc)
                  (make-application named-proc clause-exps)))
                (make-body (make-lambda ; Passed in as an argument to
                                        ; the previous lambda to
                                        ; prevent capture of clause-vars
                            clause-vars
                            body)))))
             (make-body 0)))
          (make-application
           (make-lambda clause-vars body)
           clause-exps)))))

;; Exercise 4.8:
;; Named let: to make the binding of the named procedure in scope
;; during the definition, create a place for it and bind its name to
;; that place.  This is before the procedure is defined, so set it 0
;; and use set! inside the body to set it to the required procedure.
;; Putting it in the lambda parameters ensures it is in the
;; environment.  set! sets its required value (i.e the procedure whose
;; body is that of the let body) once this is known.  Note that the
;; procedure is constructed outside the inner lambda scope to avoid
;; capturing the let parameters.

;; Worked example: fact

;; (let fact ((n 5))
;;   (if (= n 1) 1
;;       (* n (fact (- n 1)))))

;; ((lambda (fact)
;;    ((lambda (proc)
;;       (set! fact proc)
;;       (fact 5))
;;     (lambda (n)                         ; Let body
;;       (if (= n 1) 1
;;           (* n (fact (- n 1))))))
;;    0))                                  ; Placeholder value, reset by set!

;; Exercise 4.7
;; (let* ((var1 exp1) (var2 exp2) (var3 exp3)) <body>) ->
;; (let ((var1 exp1)) (let ((var2 exp2)) (let ((var3 exp3)) body)))

(define (let*->nested-let exp)
  (let ((clauses (let-clauses exp))
        (body (let-body exp)))
    (define (nested-let-from-clauses clauses)
      (make-let
       (list (let-clauses-first clauses))
       (if (let-clause-last? clauses)
           body
           (make-body (nested-let-from-clauses
                       (let-clauses-rest clauses))))))
    (nested-let-from-clauses clauses)))

;; It is sufficient to add this derived form into the interpreter to
;; transform let* into let.  This transformed expression is then
;; evaluated and transformed in turn to a combination, which is then evaluated.

;; Exercise 4.9

;; for loop
;; (for ((var1 init1 step1)
;;       (var2 init2 step2) ... )
;;   (test expression)
;;   body)
;;
;; Initialise var1 to init1, var2 to init2, ...
;; If test is true, execute body and increment init1 with step1, init2
;; with step2, ...  Otherwise return the result of evaluating expression.

(define (for? exp) (tagged-list? exp 'for))
(define (for-clauses exp) (and (for? exp) (cadr exp)))
(define for-clause-var car)
(define for-clause-init cadr)
(define for-clause-step caddr)
(define (for-test exp) (and (for? exp) (caddr exp)))
(define for-test-pred car)
(define for-test-exp cadr)
(define (for-body exp) (and (for? exp) (cdddr exp)))

(define (for->named-let exp)
  (let ((clauses (for-clauses exp))
        (test (for-test exp))
        (body (for-body exp)))
    (make-named-let
     'loop
     (map (lambda (clause)
            (list
             (for-clause-var clause)
             (for-clause-init clause)))
          clauses)
     (make-body
      (make-if (for-test-pred test)
               (make-begin
                (append body
                        (make-body
                         (make-application
                          'loop
                          (map for-clause-step clauses)))))
               (for-test-exp test))))))

;;;SECTION 4.1.3

(define (true? x)
  (not (eq? x false)))

(define (false? x)
  (eq? x false))


(define (make-procedure parameters body env)
  (list 'procedure parameters body env))

(define (compound-procedure? p)
  (tagged-list? p 'procedure))


(define (procedure-parameters p) (cadr p))
(define (procedure-body p) (caddr p))
(define (procedure-environment p) (cadddr p))


(define (enclosing-environment env) (cdr env))

(define (first-frame env) (car env))

(define the-empty-environment '())

;; Exercise 4.11: change the frame representation to be a list of
;; pairs rather than a pair of lists.
;; Frame is a headed list whose car is *frame*.  Its cdr is the first
;; backbone pair.  Each of these pairs has car of the pair (var val)
;; and cdr of the next backbone pair.  A headed list is used so that
;; new elements can be inserted while keeping the frame (i.e. the
;; first pair).
(define (make-frame variables values)
  (cons '*frame*
        (map cons variables values)))

(define (frame-bindings frame) (cdr frame))
(define (first-frame-binding frame) (cadr frame))
(define (rest-frame-bindings frame) (cddr frame))
(define (frame-variable frame-binding) (car frame-binding))
(define (frame-value frame-binding) (cdr frame-binding))
(define (frame-variables frame)
  (map frame-variable (frame-bindings frame)))
(define (frame-values frame)
  (map frame-value (frame-bindings frame)))
(define (set-frame-value! frame-binding val)
  (set-cdr! frame-binding val))

;; Assume frame is a headed list (i.e. the actual list of backbone
;; pairs is (cdr frame)). To add a binding, create a backbone pair
;; whose car is the new binding pair and whose cdr is the first
;; existing backbone pair.
(define (add-binding-to-frame! var val frame)
  (let ((binding-pair (cons var val))
        (backbone-pairs (cdr frame)))
    (set-cdr! frame
              (cons binding-pair
                    backbone-pairs))))

;; Unbind the frame binding immediately following the first one
(define (remove-next-frame-binding! frame-bindings)
              (set-cdr! frame-bindings
                        (cddr frame-bindings)))

(define (extend-environment vars vals base-env)
  (if (= (length vars) (length vals))
      (cons (make-frame vars vals) base-env)
      (if (< (length vars) (length vals))
          (error "Too many arguments supplied" vars vals)
          (error "Too few arguments supplied" vars vals))))

;; Apply a procedure to a matching variable binding. Applies
;; var-match-proc when the current binding matches the variable.
;; Applies frame-end-proc if we get to the end of the first frame
;; without a match.  If this is #f, then the search is continued in
;; subsequent frames.  Applies empty-env-proc to var if we get to the
;; end of the environment.
(define (apply-to-var var
                            var-pred
                            var-match-proc
                            frame-end-proc
                            empty-env-proc
                            env)
  (define (env-loop env)
    (define (scan frame-bindings)
      (cond ((null? frame-bindings)
             ;; If we have run out of frame bindings in this frame, we
             ;; can either go to then next frame in the environment or
             ;; run a procedure on the frame
             (if frame-end-proc
                 (frame-end-proc (first-frame env))
                 (env-loop (enclosing-environment env))))
            ((var-pred var frame-bindings)
             (var-match-proc
              frame-bindings))
            (else (scan (cdr frame-bindings)))))
    (if (eq? env the-empty-environment)
        (if empty-env-proc
            (empty-env-proc var)
            (error "Empty env -- APPLY-TO-VAR" var))
        (let ((frame (first-frame env)))
          (scan frame))))
  (env-loop env))

;; This is a bit messy: the predicate needs to check that the binding
;; is actually a pair or is the head of the list (ie just a symbol *frame*).
(define (this-frame-binding-eq? var frame-bindings)
  (and (pair? (car frame-bindings))
       (eq? var (frame-variable (car frame-bindings)))))
(define (next-frame-binding-eq? var frame-bindings)
  (and (pair? (cdr frame-bindings))
       (eq? var (frame-variable (cadr frame-bindings)))))

(define (lookup-variable-value var env)
  (apply-to-var var
                this-frame-binding-eq?
                (lambda (frame-bindings)
                  (frame-value (car frame-bindings)))
                #f
                (lambda (var)
                  (error "Variable not in environment: LOOKUP-VARIABLE-VALUE" var))
                env))

(define (set-variable-value! var val env)
  (apply-to-var var
                this-frame-binding-eq?
                (lambda (frame-bindings)
                  (set-frame-value! (car frame-bindings) val))
                #f
                (lambda (var)
                  (error "Variable not in environment: SET-VARIABLE-VALUE!" var))
                env))

(define (define-variable! var val env)
  (apply-to-var var
                this-frame-binding-eq?
                (lambda (frame-bindings)
                  (set-frame-value! (car frame-bindings) val))
                (lambda (frame)
                  (add-binding-to-frame! var val frame))
                #f
                env))

;; Exercise 4.13
;; Unbinding a variable: make-unbound-variable! should just search in
;; the current frame of the environment, as the alternative is too
;; destructive: it would affect previously formed procedures that were
;; built when the variable was bound.  This is going to need a new
;; variable match predicate, as removing an element from a linked list
;; must be done from the previous element.  So, we'll need to check
;; the next element in the list, and if it matches, then unlink the
;; next element.
(define (make-unbound-variable! var env)
  (apply-to-var var
                next-frame-binding-eq?
                (lambda (frame-bindings)
                  (remove-next-frame-binding! frame-bindings))
                (lambda (frame)
                  (error "No binding of variable in current frame" var))
                #f
                env))



;;;SECTION 4.1.4

(define (primitive-procedure? proc)
  (tagged-list? proc 'primitive))

(define (primitive-implementation proc) (cadr proc))

(define primitive-procedures
  (list (list 'car car)
        (list 'cdr cdr)
        (list 'cons cons)
        (list 'null? null?)
        (list '= =)
        (list '+ +)
        (list '* *)
        (list '- -)
        (list '/ /)
        (list '< <)
        (list '> >)
        (list 'map map)
;;      more primitives
        ))

(define (primitive-procedure-names)
  (map car
       primitive-procedures))

(define (primitive-procedure-objects)
  (map (lambda (proc) (list 'primitive (cadr proc)))
       primitive-procedures))

;[moved to start of file] (define apply-in-underlying-scheme apply)

(define (apply-primitive-procedure proc args)
  (apply
   (primitive-implementation proc) args))

(define (setup-environment)
  (let ((initial-env
         (extend-environment (primitive-procedure-names)
                             (primitive-procedure-objects)
                             the-empty-environment)))
    (define-variable! 'true true initial-env)
    (define-variable! 'false false initial-env)
    initial-env))

;[do later] (define the-global-environment (setup-environment))

(define the-global-environment (setup-environment))

(define input-prompt ";;; M-Eval input:")
(define output-prompt ";;; M-Eval value:")

(define (driver-loop)
  (prompt-for-input input-prompt)
  (let ((input (read)))
    (let ((output (eval input the-global-environment)))
      (announce-output output-prompt)
      (user-print output)))
  (driver-loop))

(define (prompt-for-input string)
  (newline) (newline) (display string) (newline))

(define (announce-output string)
  (newline) (display string) (newline))

(define (user-print object)
  (if (compound-procedure? object)
      (display (list 'compound-procedure
                     (procedure-parameters object)
                     (procedure-body object)
                     '<procedure-env>))
      (display object)))

;; Data directed version

;; Each procedure takes the expression and environment as arguments
(define dispatch-table
  (make-dispatch-table
   `((if ,eval-if)
     (begin
       ,(lambda (exp env)
          (eval-sequence (begin-actions exp) env)))
     (set! ,eval-assignment)
     (define ,eval-definition)
     (make-unbound! ,eval-unbind)
     (quote
      ,(lambda (exp env)
         (text-of-quotation exp)))
     (lambda
       ,(lambda (exp env)
          (make-procedure (lambda-parameters exp)
                          (lambda-body exp)
                          env)))
     (cond
       ,(lambda (exp env)
          (eval (cond->if exp) env)))
     (and ,eval-and)
     (or ,eval-or)
     (let ,(lambda (exp env)
             (eval (let->combination exp) env)))
     (let* ,(lambda (exp env)
              (eval (let*->nested-let exp) env)))
     (for ,(lambda (exp env)
             (eval (for->named-let exp) env))))))

(define (eval-dispatch-lookup type)
  ((dispatch-table 'lookup) type))


'METACIRCULAR-EVALUATOR-LOADED

(#%require (only racket/base module+))

(module+ main
  (define the-global-environment (setup-environment))
  ;; (driver-loop)
  )

(module+ test
  
  (define (fib n)
    (let fib-iter ((a 1)
                   (b 0)
                   (count n))
      (if (= count 0)
          b
          (fib-iter (+ a b) a (- count 1)))))

  ;; (map fib '(1 2 3 4 5 6 7 8 9 10)) ->  (1 1 2 3 5 8 13 21 34 55)
  )

;; Exercise 4.14

;; Using map from the underlying interpreter doesn't work, because the
;; interpreter and interpreted language have different representations
;; of procedures.  If we use the underlying map and pass it a
;; procedure from the interpreted language, it just sees a list
;; '(procedure ...), which is just a list of symbols in the underlying
;; lisp.

;; map definition
(define (my-map f x)
  (if (null? x)
      '()
      (cons (f (car x))
            (my-map f (cdr x)))))


